#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2014 Harpo Sp. z o. o. <http://www.harpo.com.pl>

import braille
import queueHandler
from logHandler import log
from collections import OrderedDict
from ctypes import *
from ctypes.wintypes import *
import time
import config
import inputCore
import hwPortUtils
import brailleInput
import wx
import os

file_path = os.path.dirname(os.path.abspath(__file__))

READ_INTERVAL = 50

BP12_KEYCALLBACKHANDLE = WINFUNCTYPE(None,c_ubyte,c_ubyte,c_ubyte,c_ubyte,c_ubyte,c_ubyte,c_ubyte,c_ubyte)



BP12_BRAILLE_KEYS = ("dot1", "dot2", "dot3", "dot4", "dot5", "dot6", "dot7", "dot8")
BP12_CONTROL_KEYS=("leftScroll","joyAction","joyLeft","joyUp","joyDown","joyRight","space", "rightScroll")

#Try to load BP12TI32.dll
try:
	#BP12Lib=windll[r"brailleDisplayDrivers\BP12TI.dll"]
	BP12Lib=windll.LoadLibrary(os.path.join(file_path, 'BP12TI.dll'))
except:
	BP12Lib=None

def _findPorts():
	for portInfo in hwPortUtils.listComPorts(onlyAvailable=True):
		try:
			btName = portInfo["bluetoothName"]
		except KeyError:
			continue
		if btName.startswith("EL12-"):
			yield (12, portInfo["port"])
		if btName.startswith("BP14-"):
			yield (14, portInfo["port"])
		if btName.startswith("BP"):
			pos=btName.find("-",2)
			if pos>2 and pos<5:
					bpCells = btName[2:pos]
					try:
						nbCells = int(bpCells)
						if nbCells>0:
							yield (nbCells, portInfo["port"])
					except:
						pass

def _findPortsUSB():
	for portInfo in hwPortUtils.listComPorts(onlyAvailable=True):	
		hardwareID = portInfo["hardwareID"]
		#log.info(hardwareID)
		if "VID_0483&PID_A2A8" not in hardwareID:
			continue
		yield (14, portInfo["port"])
						
class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	name = "braillePen14Driver"
	description = _("BP14, BP12 and EL12 driver")
	isThreadSafe=True

	@classmethod
	def check(cls):
		return bool(BP12Lib)
	
	@classmethod
	def getPossiblePorts(cls):
		ports = OrderedDict()
		ports["auto"] = "auto"
		for p in hwPortUtils.listComPorts():
			# Translators: Name of a serial communications port
			#log.info(p["port"])
			ports[p["port"]] = p["port"]
		return ports

#	@classmethod
#	def getManualPorts(cls):
#		return braille.getSerialPorts()

	def __init__(self, port="auto"):
		super(BrailleDisplayDriver,self).__init__()
		self.numCells = 12
		log.info("Start BP14")
		
		if port == "auto":
			portFound=False
			
			for port in _findPortsUSB():
				if not port:
					continue
				portName=u"{portNB}".format(portNB=port[1])
				self._keyCallbackInst = BP12_KEYCALLBACKHANDLE(self._keyCallback)
				result=0
				
				try:
					result=BP12Lib.open(portName,self._keyCallbackInst)
				except:
					result=0
			
				if result >0:
					self._connectedState=True
					BP12Lib.brailleDisplaySettings(port[0],0)
					self.numCells =port[0]
					portFound=True
					log.info("Found USB BP14")
					break
			
			if portFound == False:
				log.info("Search BT BP14")
				for port in _findPorts():		
					if not port:
						raise RuntimeError("No BP12 display found")
						continue
			
					portName=u"{portNB}".format(portNB=port[1])
					self._keyCallbackInst = BP12_KEYCALLBACKHANDLE(self._keyCallback)
					result=0
				
					try:
						result=BP12Lib.open(portName,self._keyCallbackInst)
					except:
						result=0
			
					if result >0:
						self._connectedState=True
						BP12Lib.brailleDisplaySettings(port[0],0)
						self.numCells =port[0]
						break
				else:
					raise RuntimeError("No BP12 display found")
		else:
			log.info("By port")
			log.info(port)
			result=0		
			conToPort=0;
			self._keyCallbackInst = BP12_KEYCALLBACKHANDLE(self._keyCallback)
			try:
				result=BP12Lib.open(port,self._keyCallbackInst)
				#log.info("Port open")
			except:
				#log.info("Open error")
				result=0
				
			if result >0:
				conToPort=1
				cde = c_ubyte();
				det = c_ubyte();
				#log.info("Get L")
				time.sleep(2.0)
				try:
					result=BP12Lib.getLength(byref(cde),byref(det))
					#log.info("Length")
					#log.info(cde.value)
					#log.info(result)
					self.numCells =cde.value
					#log.info(self.numCells)
					BP12Lib.brailleDisplaySettings(cde.value,0)
					#log.info("LengthEnd")
				except:
					#log.info("Get L Error")
					result=0
			
			if result >0:
				self._connectedState=True
				#log.info("By port OK")
			else:
				if conToPort>0:
					BP12Lib.close()
				raise RuntimeError("No BP12 display found")	
				
				

	def terminate(self):
		super(BrailleDisplayDriver, self).terminate()
		try:
			BP12Lib.close()
		except:
			pass

	def handleResponses(self):
		pass
		
	def display(self, cells):
		try:
			#arr="".join([chr(x) for x in cells])
			arr = (ctypes.c_char * len(cells))(*cells)
			if self._connectedState==True:
				result=BP12Lib.sendBraille(0,len(cells),arr)
				if result<0:
					self._connectedState=False
		except:
			pass

	def _keyCallback(self, braille, system, coursor1, coursor2, coursor3, coursor4, coursor5, coursor6):
		if braille!=0 or system!=0 or coursor1!=0:
			try:
				inputCore.manager.executeGesture(InputGesture(braille, system, coursor1))
			except inputCore.NoInputGestureAction:
				pass
				
		return
		
	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"braille_scrollBack": ("br(braillePen14Driver):leftScroll",),
			"braille_scrollForward": ("br(braillePen14Driver):rightScroll",),
			"braille_routeTo": ("br(braillePen14Driver):routing",),
			"kb:upArrow": ("br(braillePen14Driver):joyUp",),
			"kb:downArrow": ("br(braillePen14Driver):joyDown",),
			"kb:leftArrow": ("br(braillePen14Driver):joyLeft",),
			"kb:rightArrow": ("br(braillePen14Driver):joyRight",),
			"kb:enter": ("br(braillePen14Driver):joyAction","br(braillePen14Driver):space+dot8"),			
			"kb:backspace": ("br(braillePen14Driver):space+dot7","br(braillePen14Driver):space+dot1+dot2"),
			"kb:delete": ("br(braillePen14Driver):space+dot1+dot4+dot5",),
			"kb:windows": ("br(braillePen14Driver):space+dot3+dot4",),
			"kb:alt": ("br(braillePen14Driver):space+dot1+dot3+dot4",),
			"kb:escape": ("br(braillePen14Driver):space+dot1+dot5",),
			"braille_previousLine": ("br(braillePen14Driver):joyUp+space",),
			"braille_nextLine": ("br(braillePen14Driver):joyDown+space",),
			"kb:shift+tab": ("br(braillePen14Driver):space+joyLeft",),
			"kb:tab": ("br(braillePen14Driver):space+joyRight",),			
			"review_nextWord": ("br(braillePen14Driver):space+dot5",),
			"review_previousWord": ("br(braillePen14Driver):space+dot2",),
			"review_nextLine": ("br(braillePen14Driver):space+dot4",),
			"review_previousLine": ("br(braillePen14Driver):space+dot1",),
			"review_nextCharacter": ("br(braillePen14Driver):space+dot6",),
			"review_previousCharacter": ("br(braillePen14Driver):space+dot3",),
			"toggleInputHelp": ("br(braillePen14Driver):space+dot3+dot6+dot7",),
			"braille_toggleTether": ("br(braillePen14Driver):space+dot1+dot2+dot6",),
			"kb:home": ("br(braillePen14Driver):space+dot2+dot3",),
			"kb:end": ("br(braillePen14Driver):space+dot5+dot6",),
			"review_top": ("br(braillePen14Driver):space+dot1+dot2+dot3",),			
			"review_bottom": ("br(braillePen14Driver):space+dot4+dot5+dot6",),
			"sayAll": ("br(braillePen14Driver):space+dot1+dot2+dot4+dot5+dot6",),
			"review_currentCharacter": ("br(braillePen14Driver):space+dot3+dot6",),
			"review_currentWord": ("br(braillePen14Driver):space+dot2+dot5",),
			"review_currentLine": ("br(braillePen14Driver):space+dot1+dot4",),
			"dateTime": ("br(braillePen14Driver):space+dot2+dot3+dot4",),
			"reportCurrentSelection": ("br(braillePen14Driver):space+dot1+dot2+dot4+dot5+dot6+dot7",),
			"showGui": ("br(braillePen14Driver):space+dot1+dot2+dot4+dot5",),
			"kb:alt+tab": ("br(braillePen14Driver):space+dot1+dot3+dot5",),
			"say_battery_status": ("br(braillePen14Driver):space+dot1+dot2+dot7",),
			"review_markStartForCopy": ("br(braillePen14Driver):space+dot1+dot4+dot7",),
			"review_copy": ("br(braillePen14Driver):space+dot1+dot4+dot8",),
			"reportClipboardText": ("br(braillePen14Driver):space+dot1+dot2+dot3+dot6+dot8",),
			"kb:control+v": ("br(braillePen14Driver):space+dot1+dot2+dot3+dot6+dot7",),
			"review_sayAll": ("br(braillePen14Driver):space+dot1+dot3+dot4+dot5+dot6",),
			"reviewMode_next": ("br(braillePen14Driver):space+joyRight+dot7",),
			"reviewMode_previous": ("br(braillePen14Driver):space+joyLeft+dot7",),
			"review_activate": ("br(braillePen14Driver):joyAction+space",),
			"title": ("br(braillePen14Driver):space+dot2+dot3+dot4+dot5",),
			"reportCurrentFocus": ("br(braillePen14Driver):space+dot1+dot2+dot4+dot7",),
			"reportStatusLine": ("br(braillePen14Driver):space+dot2+dot3+dot4+dot7",),
			"speakForeground": ("br(braillePen14Driver):space+dot1+dot2+dot4+dot5+dot7",),
			"toggleCaretMovesReviewCursor": ("br(braillePen14Driver):joyAction+space+dot7",),
			"kb:control+leftArrow": ("br(braillePen14Driver):space+joyLeft+dot8",),
			"kb:control+rightArrow": ("br(braillePen14Driver):space+joyRight+dot8",),
			"kb:control+home": ("br(braillePen14Driver):space+joyUp+dot8",),
			"kb:control+end": ("br(braillePen14Driver):space+joyDown+dot8",),
		},
	})

class InputGesture(braille.BrailleDisplayGesture, brailleInput.BrailleInputGesture):

	source = BrailleDisplayDriver.name

	def __init__(self, brailleKeys,controlKeys,routingKey):
		super(InputGesture, self).__init__()
		
		brailleKeysList=[BP12_BRAILLE_KEYS[num] for num in range(8) if (brailleKeys>>num)&1]
		controlKeysList=[BP12_CONTROL_KEYS[num] for num in range(8) if (controlKeys>>num)&1]
		routingKeyList = ["routing"]
		
		if routingKey==0:
			selectedKeys=controlKeysList+brailleKeysList
		else:
			selectedKeys=controlKeysList+brailleKeysList+routingKeyList
		
		self.id="+".join(set(selectedKeys))
		self.keys = set(selectedKeys)
			
		
		self.keyNames = names = set()
		
		if brailleKeys!=0 and controlKeys==0 and routingKey==0:
			self.dots =brailleKeys
		
		if brailleKeys==0 and controlKeys==64 and routingKey==0:
			self.space = True
			
		if routingKey !=0:
			self.routingIndex = routingKey -1
		
		#log.info(self.id)
